# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Meanstress routines
===================

Mean stress transformation methods
----------------------------------

* FKM Goodman
* Five Segment Correction

'''

__author__ = "Johannes Mueller, Lena Rapp"
__maintainer__ = "Johannes Mueller"

import numpy as np
import pandas as pd

import pylife
from pylife.stress import stresssignal
from pylife.core import signal


@pd.api.extensions.register_dataframe_accessor("meanstress_mesh")
class MeanstressMesh(stresssignal.CyclicStressAccessor):

    def FKM_goodman(self, haigh, R_goal):
        haigh.FKM_Goodman
        Sa = self._obj.sigma_a.to_numpy()
        Sm = self._obj.sigma_m.to_numpy()
        Sa_transformed = FKM_goodman(Sa, Sm, haigh.M, haigh.M2, R_goal)
        return pd.DataFrame({'sigma_a': Sa_transformed, 'R': np.ones_like(Sa_transformed) * R_goal},
                            index=self._obj.index)

    def five_segment(self, haigh, R_goal):
        haigh.haigh_five_segment
        Sa = self._obj.sigma_a.to_numpy()
        Sm = self._obj.sigma_m.to_numpy()
        Sa_transformed = five_segment_correction(Sa, Sm,
                                                 haigh.M0, haigh.M1, haigh.M2, haigh.M3, haigh.M4,
                                                 haigh.R12, haigh.R23,
                                                 R_goal)
        return pd.DataFrame({'sigma_a': Sa_transformed, 'R': np.ones_like(Sa_transformed) * R_goal},
                            index=self._obj.index)

@pd.api.extensions.register_dataframe_accessor("meanstress_hist")
class MeanstressHist:

    def __init__(self, df):
        if df.index.names == ['from', 'to']:
            f = df.index.get_level_values('from').mid
            t = df.index.get_level_values('to').mid
            self._Sa = np.abs(f-t)/2.
            self._Sm = (f+t)/2.
            self._binsize_x = df.index.get_level_values('from').length.min()
            self._binsize_y = df.index.get_level_values('to').length.min()
        elif df.index.names == ['range', 'mean']:
            self._Sa = df.index.get_level_values('range').mid / 2.
            self._Sm = df.index.get_level_values('mean').mid
            self._binsize_x = df.index.get_level_values('range').length.min()
            self._binsize_y = df.index.get_level_values('mean').length.min()
        else:
            raise AttributeError("MeanstressHist needs index names either ['from', 'to'] or ['range', 'mean']")

        self._df = df

    def FKM_goodman(self, haigh, R_goal):
        haigh.FKM_Goodman
        Dsig = FKM_goodman(self._Sa, self._Sm, haigh.M, haigh.M2, R_goal) * 2.
        return self._rebin_results(Dsig)

    def five_segment(self, haigh, R_goal):
        haigh.haigh_five_segment
        Dsig = five_segment_correction(self._Sa, self._Sm,
                                       haigh.M0, haigh.M1, haigh.M2, haigh.M3, haigh.M4, haigh.R12,
                                       haigh.R23, R_goal) * 2.
        return self._rebin_results(Dsig)

    def _rebin_results(self, Dsig):
        Dsig_max = Dsig.max()
        binsize = np.hypot(self._binsize_x, self._binsize_y) / np.sqrt(2.)
        bincount = int(np.ceil(Dsig_max / binsize))
        new_idx = pd.IntervalIndex.from_breaks(np.linspace(0, Dsig_max, bincount), name="range")
        result = pd.DataFrame(data=np.zeros(bincount-1), index=new_idx, columns=['frequency'], dtype=np.int32)
        for i, intv in enumerate(new_idx):
            cond = np.logical_and(Dsig >= intv.left, Dsig < intv.right)
            result.loc[intv, 'frequency'] = np.int32(np.sum(self._df.values[cond]))
        result['frequency'].iloc[-1] += np.int32(np.sum(self._df.values[Dsig == Dsig_max]))
        return result

@pd.api.extensions.register_dataframe_accessor("FKM_Goodman")
@pd.api.extensions.register_series_accessor("FKM_Goodman")
class FKMGoodman:
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    def _validate(self, obj):
        signal.DataValidator().fail_if_key_missing(obj, ['M', 'M2'])


@pd.api.extensions.register_dataframe_accessor("haigh_five_segment")
@pd.api.extensions.register_series_accessor("haigh_five_segment")
class FiveSegment:
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    def _validate(self, obj):
        signal.DataValidator().fail_if_key_missing(obj, ['M0', 'M1', 'M2', 'M3', 'M4', 'R12', 'R23'])



def FKM_goodman(Sa, Sm, M, M2, R_goal):
    ''' Performs a mean stress transformation to R_goal according to the FKM-Goodman model

    :param Sa: the stress amplitude
    :param Sm: the mean stress
    :param M: the mean stress sensitivity between R=-inf and R=0
    :param M2: the mean stress sensitivity beyond R=0
    :param R_goal: the R-value to transform to

    :returns: the transformed stress range
    '''

    if R_goal == 1:
        raise ValueError('R_goal = 1 is invalid input')

    old_err_state = np.seterr(divide='ignore')

    R = np.divide(Sm-Sa, Sm+Sa)

    ignored_states = np.seterr(**old_err_state)

    c = np.where(R <= 0.)
    c2 = np.where((R > 0.) & (R < 1.))

    M = np.broadcast_to(M, Sa.shape)
    M2 = np.broadcast_to(M2, Sa.shape)

    Ma = np.zeros_like(Sa)
    Ma[c] = M[c]
    Ma[c2] = M2[c2]

    S0 = np.zeros_like(Sa)
    Sinf = np.zeros_like(Sa)

    r1 = np.where(R < 1.)
    r2 = np.where(R > 1.)

    S0[r1] = (Sa[r1]+Sm[r1]*Ma[r1])/(1.+Ma[r1])
    Sinf[r1] = S0[r1]*(1.+M[r1])/(1.-M[r1])

    Sinf[r2] = Sa[r2]
    S0[r2] = Sinf[r2]*(1.-M[r2])/(1.+M[r2])

    if R_goal == 0.0:
        return S0

    elif R_goal == -1.:
        return S0*(1.+M)

    elif R_goal == "-inf "or R_goal > 1.:
        return Sinf

    elif R_goal < 0.0:
        Mf = M

    else:
        Mf = M2

    return S0*(1.+Mf)*(1.-Mf/(Mf+(1.-R_goal)/(1.+R_goal)))


def five_segment_correction(Sa, Sm, M0, M1, M2, M3, M4, R12, R23, R_goal):
    ''' Performs a mean stress transformation to R_goal according to the
        Five Segment Mean Stress Correction

    :param Sa: the stress amplitude
    :param Sm: the mean stress
    :param Rgoal: the R-value to transform to
    :param M: the mean stress sensitivity between R=-inf and R=0
    :param M1: the mean stress sensitivity between R=0 and R=R12
    :param M2: the mean stress sensitivity betwenn R=R12 and R=R23
    :param M3: the mean stress sensitivity between R=R23 and R=1
    :param M4: the mean stress sensitivity beyond R=1
    :param R12: R-value between M1 and M2
    :param R23: R-value between M2 and M3

    :returns: the transformed stress range
    '''

    if R_goal == 1:
        raise ValueError('R_goal = 1 is invalid input')

    old_err_state = np.seterr(divide='ignore')

    R = np.divide(Sm-Sa, Sm+Sa)

    ignored_states = np.seterr(**old_err_state)

    c4 = np.where(R > 1.)
    c0 = np.where(R <= 0.)
    c1 = np.where((R > 0.) & (R <= R12))
    c2 = np.where((R > R12) & (R <= R23))
    c3 = np.where((R > R23) & (R < 1.))

    M0 = np.broadcast_to(M0, Sa.shape)
    M1 = np.broadcast_to(M1, Sa.shape)
    M2 = np.broadcast_to(M2, Sa.shape)
    M3 = np.broadcast_to(M3, Sa.shape)
    M4 = np.broadcast_to(M4, Sa.shape)

    Ma = np.zeros_like(Sa)
    Ma[c0] = M0[c0]
    Ma[c1] = M1[c1]
    Ma[c2] = M2[c2]
    Ma[c3] = M3[c3]
    Ma[c4] = M4[c4]

    S_inf = np.zeros_like(Sa)
    S_0 = np.zeros_like(Sa)
    S_12 = np.zeros_like(Sa)
    S_23 = np.zeros_like(Sa)

    B_12 = np.broadcast_to((1.+R12)/(1.-R12), Sa.shape)
    B_23 = np.broadcast_to((1.+R23)/(1.-R23), Sa.shape)

    r4 = c4
    r = np.append(c0, c1)
    r23 = np.append(c2, c3)

    S_inf[r4] = (Sa[r4]+Sm[r4]*Ma[r4])/(1.-Ma[r4])
    S_0[r4] = S_inf[r4]*(1.-M0[r4])/(1.+M0[r4])
    S_12[r4] = S_0[r4]*(1.+M1[r4])/(1.+M1[r4]*B_12[r4])
    S_23[r4] = S_12[r4]*(1.+M2[r4]*B_12[r4])/(1.+M2[r4]*B_23[r4])

    S_0[r] = (Sa[r]+Sm[r]*Ma[r])/(1.+Ma[r])
    S_inf[r] = S_0[r]*(1.+M0[r])/(1.-M0[r])
    S_12[r] = S_0[r]*(1.+M1[r])/(1.+M1[r]*B_12[r])
    S_23[r] = S_12[r]*(1.+M2[r]*B_12[r])/(1.+M2[r]*B_23[r])

    S_23[r23] = (Sa[r23]+Sm[r23]*Ma[r23])/(1.+Ma[r23]*B_23[r23])
    S_12[r23] = S_23[r23]*(1.+M2[r23]*B_23[r23])/(1.+M2[r23]*B_12[r23])
    S_0[r23] = S_12[r23]*(1.+M1[r23]*B_12[r23])/(1.+M1[r23])
    S_inf[r23] = S_0[r23]*(1.+M0[r23])/(1.-M0[r23])

    if R_goal == 0.0:
        return S_0

    if R_goal == -1.:
        return S_0*(1.+M0)

    if R_goal == -np.inf:
        return S_inf

    if R_goal == R12:
        return S_12

    if R_goal == R23:
        return S_23

    B_goal = (1.+R_goal)/(1.-R_goal)

    if R_goal <= 0.0:
        return S_0*(1.+M0)*(1.-M0/(M0+1./B_goal))

    if R_goal > 0.0 and R_goal < R12:
        return S_0*(1.+M1)*(1.-M1/(M1+1./B_goal))

    if R_goal > R12 and R_goal < R23:
        return S_23*(1.+M2*B_23)/(1.+M2*B_goal)

    if R_goal > R23 and R_goal < 1.:
        return S_23*(1.+M3*B_23)/(1.+M3*B_goal)

    if R_goal > 1.:
        return S_inf*(1.-M4)*(1.-M4/(M4+1./B_goal))


def experimental_mean_stress_sensitivity(sn_curve_R0, sn_curve_Rn1, N_c=np.inf):
    """
    Estimate the mean stress sensitivity from two `FiniteLifeCurve` objects for the same amount of cycles `N_c`.

    The formula for calculation is taken from: "Betriebsfestigkeit", Haibach, 3. Auflage 2006

    Formula (2.1-24):

    .. math::
        M_{\sigma} = {S_a}^{R=-1}(N_c) / {S_a}^{R=0}(N_c) - 1

    Alternatively the mean stress sensitivity is calculated based on both SD_50 values
    (if N_c is not given).

    Parameters
    ----------
    sn_curve_R0: pylife.strength.sn_curve.FiniteLifeCurve
        Instance of FiniteLifeCurve for R == 0
    sn_curve_Rn1: pylife.strength.sn_curve.FiniteLifeCurve
        Instance of FiniteLifeCurve for R == -1
    N_c: float, (default=np.inf)
        Amount of cycles where the amplitudes should be compared.
        If N_c is higher than a fatigue transition point (ND_50) for the SN-Curves, SD_50 is taken.
        If N_c is None, SD_50 values are taken as stress amplitudes instead.

    Returns
    -------
    float
        Mean stress sensitivity M_sigma

    Raises
    ------
    ValueError
        If the resulting M_sigma doesn't lie in the range from 0 to 1 a ValueError is raised, as this value would
        suggest higher strength with additional loads.
    """
    S_a_R0 = sn_curve_R0.calc_S(N_c) if N_c < sn_curve_R0.ND_50 else sn_curve_R0.SD_50
    S_a_Rn1 = sn_curve_Rn1.calc_S(N_c) if N_c < sn_curve_Rn1.ND_50 else sn_curve_Rn1.SD_50
    M_sigma = S_a_Rn1 / S_a_R0 - 1
    if not 0 <= M_sigma <= 1:
        raise ValueError("M_sigma: %.2f exceeds the interval [0, 1] which is not plausible." % M_sigma)
    return M_sigma



# Copyright (c) 2020-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Gyöngyvér Kiss"
__maintainer__ = __author__

from enum import Enum


class VariableLocations(Enum):
    NODE = 2
    ELEMENT_NODAL = 6


column_names = {
    'DISPLACEMENT': [['dx', 'dy', 'dz'], VariableLocations.NODE],
    'STRESS_CAUCHY': [['S11', 'S22', 'S33', 'S12', 'S13', 'S23'], VariableLocations.ELEMENT_NODAL],
    'E': [['E11', 'E22', 'E33', 'E12', 'E13', 'E23'], VariableLocations.ELEMENT_NODAL],
}



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Gyöngyvér Kiss"
__maintainer__ = __author__

from abc import ABC, abstractmethod
from IPython.display import display, clear_output
import ipywidgets as widgets


class RadioButtonWoehlerCurve(ABC):
    def __init__(self, options, description):
        self.radio_button = widgets.RadioButtons(options=options, description=description, disabled=False, style={'description_width': 'initial'})
        self.radio_button.observe(self.selection_changed_handler, names = 'value')
        display(self.radio_button)

    @abstractmethod
    def selection_changed_handler(self, change):
        pass

    def clear_selection_change_output(self):
        clear_output()
        display(self.radio_button)

    




# Copyright (c) 2020-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Gyöngyvér Kiss"
__maintainer__ = __author__

import numpy as np

from .exceptions import *
from .vmap_dataset import VMAPDataset


class VMAPCoordinateSystem(VMAPDataset):
    def __init__(self, identifier, type_id, reference_points, axis_vectors):
        super().__init__(identifier)
        self._type_id = type_id
        self._reference_points = reference_points
        self._axis_vectors = axis_vectors

    @property
    def attributes(self):
        if self._identifier is None:
            raise (APIUseError("Need to set_identifier() before requesting the attributes."))
        return self._identifier, self._type_id, self._reference_points, self._axis_vectors

    @property
    def dtype(self):
        dt_type = np.dtype({"names": ["myIdentifier", "myType", "myReferencePoint", "myAxisVectors"],
                            "formats": ['<i4', '<i4', ('<f8', (3,)), ('<f8', (9,))]})
        return dt_type

    @property
    def dataset_name(self):
        return 'COORDINATESYSTEM'




# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

from pylife.core.data_validator import DataValidator


class PylifeSignal:
    '''Base class for signal accessor classes

    Parameters
    ----------
    pandas_obj : pandas.DataFrame or pandas.Series

    Notes
    -----
    Derived classes need to implement the method `_validate(self, obj)`
    that gets `pandas_obj` as `obj` parameter. This `validate()` method
    must raise an Exception (e.g. AttributeError or ValueError) in case
    `obj` is not a valid DataFrame for the kind of signal.

    For these validation :func:`fail_if_key_missing()` and
    :func:`get_missing_keys()` might be helpful.

    For a derived class you can register methods without modifying the
    class' code itself. This can be useful if you want to make signal
    accessor classes extendable.

    See also
    --------
    :func:`fail_if_key_missing()`
    :func:`get_missing_keys()`
    :func:`register_method()`
    '''
    _method_dict = {}

    def __init__(self, pandas_obj):
        self._validator = DataValidator()
        self._validate(pandas_obj, self._validator)
        self._obj = pandas_obj

    class _MethodCaller:
        def __init__(self, method, obj):
            self._method = method
            self._obj = obj

        def __call__(self, *args, **kwargs):
            return self._method(self._obj, *args, **kwargs)

    def __getattr__(self, itemname):
        method = self._method_dict.get(itemname)

        if method is None:
            return super(PylifeSignal, self).__getattribute__(itemname)

        return self._MethodCaller(method, self._obj)

    @classmethod
    def _register_method(cls, method_name):
        def method_decorator(method):
            if method_name in cls._method_dict.keys():
                raise ValueError("Method '%s' already registered in %s" % (method_name, cls.__name__))
            if hasattr(cls, method_name):
                raise ValueError("%s already has an attribute '%s'" % (cls.__name__, method_name))
            cls._method_dict[method_name] = method
        return method_decorator


def register_method(cls, method_name):
    '''Registers a method to a class derived from :class:`PyifeSignal`

    Parameters
    ----------
    cls : class
        The class the method is registered to.
    method_name : str
        The name of the method

    Raises
    ------
    ValueError
        if `method_name` is already registered for the class
    ValueError
        if `method_name` the class has already an attribute `method_name`

    Notes
    -----
    The function is meant to be used as a decorator for a function
    that is to be installed as a method for a class. The class is
    assumed to contain a pandas object in `self._obj`.

    Examples
    --------
    .. code-block:: python

        import pandas as pd
        import pylife as pl


        @pd.api.extensions.register_dataframe_accessor('foo')
        class FooAccessor(pl.signal.PyifeSignal):
            def __init__(self, obj):
                # self._validate(obj) could come here
                self._obj = obj


        @pl.signal.register_method(FooAccessor, 'bar')
        def bar(df):
            return pd.DataFrame({'baz': df['foo'] + df['bar']})

    >>> df = pd.DataFrame({'foo': [1.0, 2.0], 'bar': [-1.0, -2.0]})
    >>> df.foo.bar()
       baz
    0  0.0
    1  0.0
    '''
    return cls._register_method(method_name)



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
import scipy.stats as stats

import pylife
from .likelihood import Likelihood
from .pearl_chain import PearlChainProbability
import pylife.utils.functions as functions
from ..accessors import FatigueDataAccessor, determine_fractures


class Elementary:
    def __init__(self, fatigue_data):
        self._fd = self._get_fatigue_data(fatigue_data)
        self._lh = self._get_likelihood()

    def _get_likelihood(self):
        return Likelihood(self._fd)

    def _get_fatigue_data(self, fatigue_data):
        if isinstance(fatigue_data, pd.DataFrame):
            if hasattr(fatigue_data, "fatigue_data"):
                params = fatigue_data.fatigue_data
            else:
                params = determine_fractures(fatigue_data).fatigue_data
        elif isinstance(fatigue_data, FatigueDataAccessor):
            params = fatigue_data
        else:
            raise ValueError("fatigue_data of type {} not understood: {}".format(type(fatigue_data), fatigue_data))
        return params

    def analyze(self, **kw):
        if len(self._fd.load.unique()) < 2:
            raise ValueError("Need at least two load levels to do a Wöhler analysis.")
        wc = self._common_analysis()
        wc = self._specific_analysis(wc, **kw)
        self.__calc_bic(wc)
        return wc

    def _common_analysis(self):
        self._slope, self._lg_intercept = self._fit_slope()
        TN_inv, TS_inv = self._pearl_chain_method()
        return pd.Series({
            'k_1': -self._slope,
            'ND_50': self._transition_cycles(self._fd.fatigue_limit),
            'SD_50': self._fd.fatigue_limit, '1/TN': TN_inv, '1/TS': TS_inv
        })

    def _specific_analysis(self, wc):
        return wc

    def bayesian_information_criterion(self):
        if not hasattr(self,"_bic"):
            raise ValueError("BIC value undefined. Analysis has not been conducted.")
        return self._bic

    def pearl_chain_estimator(self):
        return self._pearl_chain_estimator

    def __calc_bic(self, wc):
        ''' Bayesian Information Criterion: is a criterion for model selection among a finite set of models;
        the model with the lowest BIC is preferred.
        https://www.statisticshowto.datasciencecentral.com/bayesian-information-criterion/
        '''
        param_num = 5  # SD_50, 1/TS, k_1, ND_50, 1/TN
        log_likelihood = self._lh.likelihood_total(wc['SD_50'], wc['1/TS'], wc['k_1'], wc['ND_50'], wc['1/TN'])
        self._bic = (-2 * log_likelihood) + (param_num * np.log(self._fd.num_tests))

    def _fit_slope(self):
        '''Computes the slope of the finite zone with the help of a linear regression function
        '''
        slope, lg_intercept, _, _, _ = stats.linregress(np.log10(self._fd.fractures.load),
                                                        np.log10(self._fd.fractures.cycles))

        return slope, lg_intercept

    def _transition_cycles(self, fatigue_limit):
        # FIXME Elementary means fatigue_limit == 0 -> np.inf
        if fatigue_limit == 0:
            fatigue_limit = 0.1
        return 10**(self._lg_intercept + self._slope * (np.log10(fatigue_limit)))

    def _pearl_chain_method(self):
        '''
        Pearl chain method: consists of shifting the fractured data to a median load level.
        The shifted data points are assigned to a Rossow failure probability.The scatter in load-cycle
        direction can be computed from the probability net.
        '''
        self._pearl_chain_estimator = PearlChainProbability(self._fd.fractures, self._slope)

        TN_inv = functions.std2scatteringRange(1. / self._pearl_chain_estimator.slope)
        TS_inv = TN_inv**(1. / -self._slope)

        return TN_inv, TS_inv
